# -*- coding: utf-8 -*-

"""Base class of applicability domains."""


from abc import ABC, abstractmethod
from typing import Iterable, Union

from sklearn.utils.validation import check_array, column_or_1d

from .utils import check_is_fitted


class ApplicabilityDomain(ABC):
    """Base class applicability domains inherits from."""

    def __init__(self):
        self.fitted_ = False

    def fit(self, X):
        """Fit the applicability domain to the given feature matrix

        :param X: feature matrix
        """
        # Ensure is an array and is finite
        X = check_array(X)
        self.num_points, self.num_dims = X.shape
        # Call underlying _fit method
        self._fit(X)
        self.fitted_ = True

    @abstractmethod
    def _fit(self, X):
        """Method to overload to fit the model

        :param X: feature matrix
        """
        pass

    def contains(self, sample) -> Union[bool, Iterable[bool]]:
        """Determine if a sample is in the applicability domain.

        :param sample: sample to check the applicability domain membership of.
        """
        # Ensure the AD was fitted
        check_is_fitted(self, 'fitted_')
        try:
            # Ensure is 1D sample
            sample = column_or_1d(sample)
        except ValueError:
            sample = check_array(sample, accept_large_sparse=False)
        # Check dimensions
        if sample.ndim == 1 and sample.shape[0] != self.num_dims:
            raise ValueError('sample must have the same number of features as the applicability domain; '
                             f'{sample.shape[0]} and {self.num_dims} respectively')
        elif sample.ndim == 2 and sample.shape[1] != self.num_dims:
            raise ValueError('sample must have the same number of features as the applicability domain; '
                             f'{sample.shape[1]} and {self.num_dims} respectively')
        return self._contains(sample)

    @abstractmethod
    def _contains(self, sample):
        """Method to overload to determine if a sample is in the applicability domain.

        :param sample: sample to check the applicability domain membership of.
        """
        pass



# -*- coding: utf-8 -*-

import unittest

from sklearn.datasets import make_blobs

from src.mlchemad import *


class TestAD(unittest.TestCase):

    def setUp(self):
        self.X_cent1_sd1, _ = make_blobs(n_samples=1000, n_features=100, centers=[[1] * 100], cluster_std=1, random_state=1234)
        self.X_cent1_sd3, _ = make_blobs(n_samples=1000, n_features=100, centers=[[1] * 100], cluster_std=3, random_state=1234)
        self.X_cent6_sd1, _ = make_blobs(n_samples=1000, n_features=100, centers=[[6] * 100], cluster_std=1, random_state=1234)
        self.X_cent6_sd3, _ = make_blobs(n_samples=1000, n_features=100, centers=[[6] * 100], cluster_std=3, random_state=1234)
        self.mekenyan_veith = data

    def test_minmax_boundingbox(self):
        ad = BoundingBoxApplicabilityDomain(percentiles=None)
        ad.fit(self.X_cent1_sd1)
        self.assertEquals(sum(ad.contains(self.X_cent1_sd1)),
                          len(self.X_cent1_sd1))
        self.assertEquals(sum(ad.contains(self.X_cent6_sd1)),
                          0)
        self.assertLess(sum(ad.contains(self.X_cent1_sd3)),
                        len(self.X_cent1_sd3))
        self.assertEquals(ad.contains(self.X_cent1_sd1).tolist(),
                          [ad.contains(x) for x in self.X_cent1_sd1])
        self.assertEquals(ad.contains(self.X_cent6_sd1).tolist(),
                          [ad.contains(x) for x in self.X_cent6_sd1])
        self.assertEquals(ad.contains(self.X_cent1_sd3).tolist(),
                          [ad.contains(x) for x in self.X_cent1_sd3])
        ad.fit(self.mekenyan_veith.training)
        self.assertGreaterEqual(sum(ad.contains(self.mekenyan_veith.test)), 0)
        self.assertLessEqual(sum(ad.contains(self.mekenyan_veith.test)), self.mekenyan_veith.test.shape[0])


    def test_fixed_minmax_boundingbox(self):
        ad = BoundingBoxApplicabilityDomain(percentiles=None, range_=(-3, 5))
        ad.fit(self.X_cent1_sd1)
        self.assertGreater(sum(ad.contains(self.X_cent1_sd1)),
                           0.9 * len(self.X_cent1_sd1))
        self.assertEquals(sum(ad.contains(self.X_cent6_sd1)),
                          0)
        self.assertLess(sum(ad.contains(self.X_cent1_sd3)),
                        len(self.X_cent1_sd3))
        self.assertEquals(ad.contains(self.X_cent1_sd1).tolist(),
                          [ad.contains(x) for x in self.X_cent1_sd1])
        self.assertEquals(ad.contains(self.X_cent6_sd1).tolist(),
                          [ad.contains(x) for x in self.X_cent6_sd1])
        self.assertEquals(ad.contains(self.X_cent1_sd3).tolist(),
                          [ad.contains(x) for x in self.X_cent1_sd3])
        ad.fit(self.mekenyan_veith.training)
        self.assertGreaterEqual(sum(ad.contains(self.mekenyan_veith.test)), 0)
        self.assertLessEqual(sum(ad.contains(self.mekenyan_veith.test)), self.mekenyan_veith.test.shape[0])

    def test_percentiles_boundingbox(self):
        ad = BoundingBoxApplicabilityDomain()
        ad.fit(self.X_cent1_sd1)
        self.assertGreater(sum(ad.contains(self.X_cent1_sd1)),
                           0.8 * len(self.X_cent1_sd1))
        self.assertEquals(sum(ad.contains(self.X_cent6_sd1)),
                          0)
        self.assertLess(sum(ad.contains(self.X_cent1_sd3)),
                        len(self.X_cent1_sd3))
        self.assertEquals(ad.contains(self.X_cent1_sd1).tolist(),
                          [ad.contains(x) for x in self.X_cent1_sd1])
        self.assertEquals(ad.contains(self.X_cent6_sd1).tolist(),
                          [ad.contains(x) for x in self.X_cent6_sd1])
        self.assertEquals(ad.contains(self.X_cent1_sd3).tolist(),
                          [ad.contains(x) for x in self.X_cent1_sd3])
        ad.fit(self.mekenyan_veith.training)
        self.assertGreaterEqual(sum(ad.contains(self.mekenyan_veith.test)), 0)
        self.assertLessEqual(sum(ad.contains(self.mekenyan_veith.test)), self.mekenyan_veith.test.shape[0])

    def test_convexhull(self):
        ad = ConvexHullApplicabilityDomain()
        ad.fit(self.X_cent1_sd1)
        self.assertEquals(ad.contains(self.X_cent1_sd1[:100]).tolist(),
                          [ad.contains(x) for x in self.X_cent1_sd1[:100]])
        self.assertEquals(ad.contains(self.X_cent6_sd1[:100]).tolist(),
                          [ad.contains(x) for x in self.X_cent6_sd1[:100]])
        self.assertEquals(ad.contains(self.X_cent1_sd3[:100]).tolist(),
                          [ad.contains(x) for x in self.X_cent1_sd3[:100]])
        ad.fit(self.mekenyan_veith.training)
        self.assertGreaterEqual(sum(ad.contains(self.mekenyan_veith.test)), 0)
        self.assertLessEqual(sum(ad.contains(self.mekenyan_veith.test)), self.mekenyan_veith.test.shape[0])

    def test_pca_robust_boundingbox(self):
        ad = PCABoundingBoxApplicabilityDomain()
        ad.fit(self.X_cent1_sd1)
        self.assertGreater(sum(ad.contains(self.X_cent1_sd1)),
                           0.9 * len(self.X_cent1_sd1))
        self.assertEquals(sum(ad.contains(self.X_cent6_sd1)),
                          0)
        self.assertLess(sum(ad.contains(self.X_cent1_sd3)),
                        len(self.X_cent1_sd3))
        ad.fit(self.mekenyan_veith.training)
        self.assertGreaterEqual(sum(ad.contains(self.mekenyan_veith.test)), 0)
        self.assertLessEqual(sum(ad.contains(self.mekenyan_veith.test)), self.mekenyan_veith.test.shape[0])

    def test_pca_standard_boundingbox(self):
        ad = PCABoundingBoxApplicabilityDomain(scaling='standard')
        ad.fit(self.X_cent1_sd1)
        self.assertGreater(sum(ad.contains(self.X_cent1_sd1)),
                           0.9 * len(self.X_cent1_sd1))
        self.assertEquals(sum(ad.contains(self.X_cent6_sd1)),
                          0)
        self.assertLess(sum(ad.contains(self.X_cent1_sd3)),
                        len(self.X_cent1_sd3))
        ad.fit(self.mekenyan_veith.training)
        self.assertGreaterEqual(sum(ad.contains(self.mekenyan_veith.test)), 0)
        self.assertLessEqual(sum(ad.contains(self.mekenyan_veith.test)), self.mekenyan_veith.test.shape[0])

    def test_pca_minmax_boundingbox(self):
        ad = PCABoundingBoxApplicabilityDomain(scaling='minmax')
        ad.fit(self.X_cent1_sd1)
        self.assertGreater(sum(ad.contains(self.X_cent1_sd1)),
                          0.9 * len(self.X_cent1_sd1))
        self.assertEquals(sum(ad.contains(self.X_cent6_sd1)),
                          0)
        self.assertLess(sum(ad.contains(self.X_cent1_sd3)),
                        len(self.X_cent1_sd3))
        ad.fit(self.mekenyan_veith.training)
        self.assertGreaterEqual(sum(ad.contains(self.mekenyan_veith.test)), 0)
        self.assertLessEqual(sum(ad.contains(self.mekenyan_veith.test)), self.mekenyan_veith.test.shape[0])

    def test_pca_maxabs_boundingbox(self):
        ad = PCABoundingBoxApplicabilityDomain(scaling='maxabs')
        ad.fit(self.X_cent1_sd1)
        self.assertGreater(sum(ad.contains(self.X_cent1_sd1)),
                           0.9 * len(self.X_cent1_sd1))
        self.assertEquals(sum(ad.contains(self.X_cent6_sd1)),
                          0)
        self.assertLess(sum(ad.contains(self.X_cent1_sd3)),
                        len(self.X_cent1_sd3))
        ad.fit(self.mekenyan_veith.training)
        self.assertGreaterEqual(sum(ad.contains(self.mekenyan_veith.test)), 0)
        self.assertLessEqual(sum(ad.contains(self.mekenyan_veith.test)), self.mekenyan_veith.test.shape[0])

    def test_leverage(self):
        ad = LeverageApplicabilityDomain()
        ad.fit(self.X_cent1_sd1)
        self.assertGreater(sum(ad.contains(self.X_cent1_sd1)),
                           0.9 * len(self.X_cent1_sd1))
        self.assertEquals(sum(ad.contains(self.X_cent6_sd3)),
                          0)
        self.assertLess(sum(ad.contains(self.X_cent1_sd3)),
                        len(self.X_cent1_sd3))
        self.assertEquals(ad.contains(self.X_cent1_sd1).tolist(),
                          [ad.contains(x) for x in self.X_cent1_sd1])
        self.assertEquals(ad.contains(self.X_cent6_sd1).tolist(),
                          [ad.contains(x) for x in self.X_cent6_sd1])
        self.assertEquals(ad.contains(self.X_cent1_sd3).tolist(),
                          [ad.contains(x) for x in self.X_cent1_sd3])
        ad.fit(self.mekenyan_veith.training)
        self.assertGreaterEqual(sum(ad.contains(self.mekenyan_veith.test)), 0)
        self.assertLessEqual(sum(ad.contains(self.mekenyan_veith.test)), self.mekenyan_veith.test.shape[0])

    def test_hotellingt2(self):
        ad = HotellingT2ApplicabilityDomain()
        ad.fit(self.X_cent1_sd1)
        self.assertGreater(sum(ad.contains(self.X_cent1_sd1)),
                           0.9 * len(self.X_cent1_sd1))
        self.assertEquals(sum(ad.contains(self.X_cent6_sd3)),
                          0)
        self.assertLess(sum(ad.contains(self.X_cent1_sd3)),
                        len(self.X_cent1_sd3))
        self.assertEquals(ad.contains(self.X_cent1_sd1).tolist(),
                          [ad.contains(x) for x in self.X_cent1_sd1])
        self.assertEquals(ad.contains(self.X_cent6_sd1).tolist(),
                          [ad.contains(x) for x in self.X_cent6_sd1])
        self.assertEquals(ad.contains(self.X_cent1_sd3).tolist(),
                          [ad.contains(x) for x in self.X_cent1_sd3])
        ad.fit(self.mekenyan_veith.training)
        self.assertGreaterEqual(sum(ad.contains(self.mekenyan_veith.test)), 0)
        self.assertLessEqual(sum(ad.contains(self.mekenyan_veith.test)), self.mekenyan_veith.test.shape[0])

    def test_centroid_distance(self):
        ad = CentroidDistanceApplicabilityDomain()
        ad.fit(self.X_cent1_sd1)
        self.assertGreater(sum(ad.contains(self.X_cent1_sd1)),
                           0.9 * len(self.X_cent1_sd1))
        self.assertEquals(sum(ad.contains(self.X_cent6_sd1)),
                          0)
        self.assertLess(sum(ad.contains(self.X_cent1_sd3)),
                        len(self.X_cent1_sd3))
        self.assertEquals(ad.contains(self.X_cent1_sd1).tolist(),
                          [ad.contains(x) for x in self.X_cent1_sd1])
        self.assertEquals(ad.contains(self.X_cent6_sd1).tolist(),
                          [ad.contains(x) for x in self.X_cent6_sd1])
        self.assertEquals(ad.contains(self.X_cent1_sd3).tolist(),
                          [ad.contains(x) for x in self.X_cent1_sd3])
        ad.fit(self.mekenyan_veith.training)
        self.assertGreaterEqual(sum(ad.contains(self.mekenyan_veith.test)), 0)
        self.assertLessEqual(sum(ad.contains(self.mekenyan_veith.test)), self.mekenyan_veith.test.shape[0])

    def test_knn(self):
        ad = KNNApplicabilityDomain()
        ad.fit(self.X_cent1_sd1)
        self.assertGreater(sum(ad.contains(self.X_cent1_sd1)),
                           0.9 * len(self.X_cent1_sd1))
        self.assertEquals(sum(ad.contains(self.X_cent6_sd1)),
                          0)
        self.assertLess(sum(ad.contains(self.X_cent1_sd3)),
                        len(self.X_cent1_sd3))

        self.assertEquals(ad.contains(self.X_cent1_sd1).tolist(),
                          [ad.contains(x) for x in self.X_cent1_sd1])
        self.assertEquals(ad.contains(self.X_cent6_sd1).tolist(),
                          [ad.contains(x) for x in self.X_cent6_sd1])
        self.assertEquals(ad.contains(self.X_cent1_sd3).tolist(),
                          [ad.contains(x) for x in self.X_cent1_sd3])
        self.assertEquals(ad.contains(self.X_cent1_sd1).tolist(),
                          [ad.contains(x) for x in self.X_cent1_sd1])
        self.assertEquals(ad.contains(self.X_cent6_sd1).tolist(),
                          [ad.contains(x) for x in self.X_cent6_sd1])
        self.assertEquals(ad.contains(self.X_cent1_sd3).tolist(),
                          [ad.contains(x) for x in self.X_cent1_sd3])
        ad.fit(self.mekenyan_veith.training)
        self.assertGreaterEqual(sum(ad.contains(self.mekenyan_veith.test)), 0)
        self.assertLessEqual(sum(ad.contains(self.mekenyan_veith.test)), self.mekenyan_veith.test.shape[0])

    def test_kerneldensity(self):
        ad = KernelDensityApplicabilityDomain()
        ad.fit(self.X_cent1_sd1)
        self.assertGreater(sum(ad.contains(self.X_cent1_sd1)),
                           0.9 * len(self.X_cent1_sd1))
        self.assertEquals(sum(ad.contains(self.X_cent6_sd1)),
                          0)
        self.assertLess(sum(ad.contains(self.X_cent1_sd3)),
                        len(self.X_cent1_sd3))
        self.assertEquals(ad.contains(self.X_cent1_sd1).tolist(),
                          [ad.contains(x) for x in self.X_cent1_sd1])
        self.assertEquals(ad.contains(self.X_cent6_sd1).tolist(),
                          [ad.contains(x) for x in self.X_cent6_sd1])
        self.assertEquals(ad.contains(self.X_cent1_sd3).tolist(),
                          [ad.contains(x) for x in self.X_cent1_sd3])
        ad.fit(self.mekenyan_veith.training)
        self.assertGreaterEqual(sum(ad.contains(self.mekenyan_veith.test)), 0)
        self.assertLessEqual(sum(ad.contains(self.mekenyan_veith.test)), self.mekenyan_veith.test.shape[0])

    def test_topkat(self):
        ad = TopKatApplicabilityDomain()
        ad.fit(self.X_cent1_sd1)
        self.assertEquals(sum(ad.contains(self.X_cent1_sd1)),
                          len(self.X_cent1_sd1))
        self.assertEquals(sum(ad.contains(self.X_cent6_sd1)),
                          0)
        self.assertLess(sum(ad.contains(self.X_cent1_sd3)),
                        len(self.X_cent1_sd3))
        self.assertEquals(ad.contains(self.X_cent1_sd1).tolist(),
                          [ad.contains(x) for x in self.X_cent1_sd1])
        self.assertEquals(ad.contains(self.X_cent6_sd1).tolist(),
                          [ad.contains(x) for x in self.X_cent6_sd1])
        self.assertEquals(ad.contains(self.X_cent1_sd3).tolist(),
                          [ad.contains(x) for x in self.X_cent1_sd3])
        ad.fit(self.mekenyan_veith.training)
        self.assertGreaterEqual(sum(ad.contains(self.mekenyan_veith.test)), 0)
        self.assertLessEqual(sum(ad.contains(self.mekenyan_veith.test)), self.mekenyan_veith.test.shape[0])

    def test_isolation_forest(self):
        ad = IsolationForestApplicabilityDomain()
        ad.fit(self.X_cent1_sd1)
        self.assertGreater(sum(ad.contains(self.X_cent1_sd1)),
                           0.9 * len(self.X_cent1_sd1))
        self.assertEquals(sum(ad.contains(self.X_cent6_sd1)),
                          0)
        self.assertLess(sum(ad.contains(self.X_cent1_sd3)),
                        len(self.X_cent1_sd3))
        self.assertEquals(ad.contains(self.X_cent1_sd1).tolist(),
                          [ad.contains(x) for x in self.X_cent1_sd1])
        self.assertEquals(ad.contains(self.X_cent6_sd1).tolist(),
                          [ad.contains(x) for x in self.X_cent6_sd1])
        self.assertEquals(ad.contains(self.X_cent1_sd3).tolist(),
                          [ad.contains(x) for x in self.X_cent1_sd3])
        ad.fit(self.mekenyan_veith.training)
        self.assertGreaterEqual(sum(ad.contains(self.mekenyan_veith.test)), 0)
        self.assertLessEqual(sum(ad.contains(self.mekenyan_veith.test)), self.mekenyan_veith.test.shape[0])



# -*- coding: utf-8 -*-

"""Toy data."""


import os
import json

import pandas as pd


def _read_data():
    # Obtain a file handle
    with open(os.path.abspath(os.path.join(__file__, os.pardir, 'data.json'))) as fh:
        toy_data = json.load(fh)
    return (toy_data.get('source'),
            pd.DataFrame(toy_data.get('training')).set_index('Compound'),
            pd.DataFrame(toy_data.get('test')).set_index('Compound')
            )


source, training, test = _read_data()
